XX(uc_version)
XX(uc_arch_supported)
XX(uc_open)
XX(uc_close)
XX(uc_query)
XX(uc_ctl)
XX(uc_errno)
XX(uc_strerror)
XX(uc_reg_write)
XX(uc_reg_read)
XX(uc_reg_write2)
XX(uc_reg_read2)
XX(uc_reg_write_batch)
XX(uc_reg_read_batch)
XX(uc_reg_write_batch2)
XX(uc_reg_read_batch2)
XX(uc_mem_write)
XX(uc_mem_read)
XX(uc_vmem_read)
XX(uc_vmem_write)
XX(uc_vmem_translate)
XX(uc_emu_start)
XX(uc_emu_stop)
XX(uc_hook_add)
XX(uc_hook_del)
XX(uc_mem_map)
XX(uc_mem_map_ptr)
XX(uc_mmio_map)
XX(uc_mem_unmap)
XX(uc_mem_protect)
XX(uc_mem_regions)
XX(uc_context_alloc)
XX(uc_free)
XX(uc_context_save)
XX(uc_context_reg_write)
XX(uc_context_reg_read)
XX(uc_context_reg_write2)
XX(uc_context_reg_read2)
XX(uc_context_reg_write_batch)
XX(uc_context_reg_read_batch)
XX(uc_context_reg_write_batch2)
XX(uc_context_reg_read_batch2)
XX(uc_context_restore)
XX(uc_context_size)
XX(uc_context_free)
